"""Parser implementations for disnake emoji types."""

from __future__ import annotations

import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers
from disnake_compass.internal import di

__all__: typing.Sequence[str] = (
    "EmojiParser",
    "PartialEmojiParser",
    "StickerParser",
)


# GET_ONLY


# TODO: Probably need to implement animated, maybe also name
# TODO: Maybe implement some way of *not* requiring ids for partial emoji
@parser_base.register_parser_for(disnake.PartialEmoji)
@attrs.define(slots=True)
class PartialEmojiParser(parser_base.Parser[disnake.PartialEmoji]):
    r"""Parser type with support for partial emoji.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default partial emoji parser will also return compressed results.
    """

    async def loads(self, argument: str, /) -> disnake.PartialEmoji:
        """Load a partial emoji from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a partial emoji.

        """
        return disnake.PartialEmoji.from_dict({"id": self.int_parser.loads(argument)})

    async def dumps(self, argument: disnake.PartialEmoji, /) -> str:
        """Dump a partial emoji into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        if not argument.id:
            msg = "PartialEmojiParser requires PartialEmoji.id to be set."
            raise ValueError(msg)

        return await self.int_parser.dumps(argument.id)


@parser_base.register_parser_for(disnake.Emoji)
@attrs.define(slots=True, kw_only=True)
class EmojiParser(parser_base.Parser[disnake.Emoji]):
    """Synchronous parser type with support for emoji.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default emoji parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.Emoji:
        """Load an emoji from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into an emoji.

        Raises
        ------
        :class:`LookupError`:
            An emoji with the id stored in the ``argument`` could not be found.

        """
        emoji_id = await self.int_parser.loads(argument)
        client = di.resolve_dependency(disnake.Client)

        emoji = client.get_emoji(emoji_id)
        if emoji:
            return emoji

        if self.allow_api_requests:
            guild = di.resolve_dependency(disnake.Guild)
            return await guild.fetch_emoji(emoji_id)

        msg = f"Could not find an emoji with id {emoji_id}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Emoji, /) -> str:
        """Dump an emoji into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)


@parser_base.register_parser_for(disnake.Sticker)
@attrs.define(slots=True, kw_only=True)
class StickerParser(parser_base.Parser[disnake.Sticker]):
    """Synchronous parser type with support for stickers.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default emoji parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.Sticker:
        """Load a sticker from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a sticker.

        Raises
        ------
        :class:`LookupError`:
            A sticker with the id stored in the ``argument`` could not be found.

        """
        client = di.resolve_dependency(disnake.Client)
        sticker_id = await self.int_parser.loads(argument)

        sticker = client.get_sticker(sticker_id)
        if sticker:
            return sticker

        if self.allow_api_requests:
            guild = di.resolve_dependency(disnake.Guild)
            return await guild.fetch_sticker(sticker_id)

        msg = f"Could not find an emoji with id {sticker_id}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Sticker, /) -> str:
        """Dump a sticker into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)



"""Parser implementations for disnake.Guild type."""

from __future__ import annotations

import contextlib
import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers
from disnake_compass.internal import di

__all__: typing.Sequence[str] = (
    "GuildParser",
    "InviteParser",
    "RoleParser",
)


@parser_base.register_parser_for(disnake.Guild)
@attrs.define(slots=True)
class GuildParser(parser_base.Parser[disnake.Guild]):
    """Parser type with support for guilds.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.Guild:
        """Load a guild from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a guild.

        Raises
        ------
        :class:`LookupError`:
            A guild with the id stored in the ``argument`` could not be found.

        """
        guild_id = await self.int_parser.loads(argument)

        maybe_guild = di.resolve_dependency(disnake.Guild, None)
        if maybe_guild and maybe_guild.id == guild_id:
            return maybe_guild

        maybe_client = di.resolve_dependency(disnake.Client, None)
        if maybe_client:
            maybe_guild = maybe_client.get_guild(guild_id)
            if maybe_guild:
                return maybe_guild

            if self.allow_api_requests:
                with contextlib.suppress(disnake.HTTPException):
                    return await maybe_client.fetch_guild(guild_id)

        msg = f"Could not find a guild with id {guild_id}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Guild) -> str:
        """Dump a guild into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)


@parser_base.register_parser_for(disnake.Invite)
@attrs.define(slots=True, kw_only=True)
class InviteParser(parser_base.Parser[disnake.Invite]):
    """Parser type with support for guilds.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    with_counts:
        Whether to include count information in the invite.
    with_expiration:
        Whether to include the expiration date of the invite.
    guild_scheduled_event_id: :class:`int`
        The ID of the scheduled event to include in the invite.

        If not provided, defaults to the ``event`` parameter in the URL if
        it exists, or the ID of the scheduled event contained in the
        provided invite object.

    """

    with_counts: bool = attrs.field(default=True)
    """Whether to include the number of times an invite was used."""
    with_expiration: bool = attrs.field(default=True)
    """Whether to include when the invite expires."""
    guild_scheduled_event_id: int | None = attrs.field(default=None)
    """The ID of the scheduled event to include in the invite."""

    async def loads(self, argument: str) -> disnake.Invite:
        """Load a guild invite from a string.

        This uses the underlying :attr:`int_parser`.

        This method first tries to get the invite from cache. If this fails,
        it will try to fetch the invite instead.

        .. warning::
            This method can make API requests.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a guild invite.

        """
        client = di.resolve_dependency(disnake.Client)
        return await client.fetch_invite(
            argument,
            with_counts=self.with_counts,
            with_expiration=self.with_expiration,
            guild_scheduled_event_id=self.guild_scheduled_event_id,
        )

    async def dumps(self, argument: disnake.Invite, /) -> str:
        """Dump a guild invite into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return argument.id


@parser_base.register_parser_for(disnake.Role)
@attrs.define(slots=True)
class RoleParser(parser_base.Parser[disnake.Role]):
    r"""Parser type with support for roles.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str) -> disnake.Role:
        """Load a role from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a role.

        Raises
        ------
        :class:`LookupError`:
            A role with the id stored in the ``argument`` could not be found.

        """
        role_id = await self.int_parser.loads(argument)

        guild = di.resolve_dependency(disnake.Guild)
        role = guild.get_role(role_id)
        if role:
            return role

        if self.allow_api_requests:
            with contextlib.suppress(disnake.HTTPException):
                for role in await guild.fetch_roles():
                    if role.id == role_id:
                        return role

        msg = f"Could not find a role with id {argument!r}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Role) -> str:
        """Dump a role into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)



"""Default implementation of select-based components."""

from __future__ import annotations

import typing

import attrs
import disnake

from disnake_compass import fields
from disnake_compass.api import component as component_api
from disnake_compass.impl.component import base as component_base

__all__: typing.Sequence[str] = (
    "RichChannelSelect",
    "RichMentionableSelect",
    "RichRoleSelect",
    "RichStringSelect",
    "RichUserSelect",
)


class BaseSelect(
    component_api.RichSelect,
    component_base.ComponentBase,
    typing.Protocol,
):
    """The base class of a disnake-compass select menu.

    For implementations, see :class:`RichStringSelect`, :class:`RichUserSelect`,
    :class:`RichRoleSelect`, :class:`RichMentionableSelect`,
    :class:`RichChannelSelect`.
    """

    event: typing.ClassVar[str] = "on_dropdown"

    placeholder: str | None = fields.internal(default=None)
    min_values: int = fields.internal(default=1)
    max_values: int = fields.internal(default=1)
    disabled: bool = fields.internal(default=False)


class RichStringSelect(BaseSelect, typing.Protocol):
    """The default implementation of a disnake-compass string select.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`placeholder`,
    :attr:`min_values`, :attr:`max_values`, :attr:`disabled`, and:attr:`options`.
    These set the corresponding attributes on the select class when they are
    sent to discord, and are meant to be overwritten by the user.

    Fields can be defined similarly to dataclasses, by means of a name, a type
    annotation, and an optional :func:`disnake_compass.field` to set the default or
    a custom parser. The options field specifically is designated with
    :func:`disnake_compass.options` instead.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    options: list[disnake.SelectOption] = fields.internal(
        default=attrs.Factory(list[disnake.SelectOption]),
    )
    """The options for this select menu.

    Must be a list of between 1 and 25 strings.
    """

    async def as_ui_component(self) -> disnake.ui.StringSelect[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.StringSelect(
            placeholder=self.placeholder,
            min_values=self.min_values,
            max_values=self.max_values,
            disabled=self.disabled,
            options=self.options,
            custom_id=await self.manager.make_custom_id(self),
        )


class RichUserSelect(BaseSelect, typing.Protocol):
    """The default implementation of a disnake-compass user select.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`placeholder`,
    :attr:`min_values`, :attr:`max_values`, :attr:`disabled`.
    These set the corresponding attributes on the select class when they are
    sent to discord, and are meant to be overwritten by the user.

    Fields can be defined similarly to dataclasses, by means of a name, a type
    annotation, and an optional :func:`disnake_compass.field` to set the default or
    a custom parser. The options field specifically is designated with
    :func:`disnake_compass.options` instead.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    async def as_ui_component(self) -> disnake.ui.UserSelect[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.UserSelect(
            placeholder=self.placeholder,
            min_values=self.min_values,
            max_values=self.max_values,
            disabled=self.disabled,
            custom_id=await self.manager.make_custom_id(self),
        )


class RichRoleSelect(BaseSelect, typing.Protocol):
    """The default implementation of a disnake-compass role select.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`placeholder`,
    :attr:`min_values`, :attr:`max_values`, :attr:`disabled`.
    These set the corresponding attributes on the select class when they are
    sent to discord, and are meant to be overwritten by the user.

    Fields can be defined similarly to dataclasses, by means of a name, a type
    annotation, and an optional :func:`disnake_compass.field` to set the default or
    a custom parser. The options field specifically is designated with
    :func:`disnake_compass.options` instead.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    async def as_ui_component(self) -> disnake.ui.RoleSelect[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.RoleSelect(
            placeholder=self.placeholder,
            min_values=self.min_values,
            max_values=self.max_values,
            disabled=self.disabled,
            custom_id=await self.manager.make_custom_id(self),
        )


class RichMentionableSelect(BaseSelect, typing.Protocol):
    """The default implementation of a disnake-compass mentionable select.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`placeholder`,
    :attr:`min_values`, :attr:`max_values`, :attr:`disabled`.
    These set the corresponding attributes on the select class when they are
    sent to discord, and are meant to be overwritten by the user.

    Fields can be defined similarly to dataclasses, by means of a name, a type
    annotation, and an optional :func:`disnake_compass.field` to set the default or
    a custom parser. The options field specifically is designated with
    :func:`disnake_compass.options` instead.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    async def as_ui_component(self) -> disnake.ui.MentionableSelect[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.MentionableSelect(
            placeholder=self.placeholder,
            min_values=self.min_values,
            max_values=self.max_values,
            disabled=self.disabled,
            custom_id=await self.manager.make_custom_id(self),
        )


class RichChannelSelect(BaseSelect, typing.Protocol):
    """The default implementation of a disnake-compass channel select.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`channel_types`,
    :attr:`placeholder`, :attr:`min_values`, :attr:`max_values`, :attr:`disabled`.
    These set the corresponding attributes on the select class when they are
    sent to discord, and are meant to be overwritten by the user.

    Fields can be defined similarly to dataclasses, by means of a name, a type
    annotation, and an optional :func:`disnake_compass.field` to set the default or
    a custom parser. The options field specifically is designated with
    :func:`disnake_compass.options` instead.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    channel_types: list[disnake.ChannelType] | None = fields.internal(
        default=None,
    )
    """The channel types to allow for this select menu.

    Defaults to :obj:`None`, implying all channel types are allowed.
    """

    async def as_ui_component(self) -> disnake.ui.ChannelSelect[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.ChannelSelect(
            channel_types=self.channel_types,
            placeholder=self.placeholder,
            min_values=self.min_values,
            max_values=self.max_values,
            disabled=self.disabled,
            custom_id=await self.manager.make_custom_id(self),
        )



"""Parser implementations for disnake user types."""

from __future__ import annotations

import contextlib
import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers
from disnake_compass.internal import di

__all__: typing.Sequence[str] = ("MemberParser", "UserParser")


@parser_base.register_parser_for(disnake.User)
@attrs.define(slots=True)
class UserParser(parser_base.Parser[disnake.User]):
    r"""Parser type with support for users.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.User:
        """Load a user from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a user.

        Raises
        ------
        :class:`LookupError`:
            A user with the id stored in the ``argument`` could not be found.

        """
        user_id = await self.int_parser.loads(argument)

        maybe_author = di.resolve_dependency(disnake.member._UserTag, None)  # pyright: ignore[reportPrivateUsage, reportPrivateImportUsage]  # noqa: SLF001
        if maybe_author and maybe_author.id == user_id:
            if isinstance(maybe_author, disnake.User):
                return maybe_author
            if isinstance(maybe_author, disnake.Member):
                return maybe_author._user  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001

        maybe_client = di.resolve_dependency(disnake.Client, None)
        if maybe_client:
            user = maybe_client.get_user(user_id)
            if user:
                return user

            if self.allow_api_requests:
                with contextlib.suppress(disnake.HTTPException):
                    return await maybe_client.fetch_user(user_id)

        msg = f"Could not find a user with id {argument!r}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.User, /) -> str:
        """Dump a user into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)


@parser_base.register_parser_for(disnake.Member)
@attrs.define(slots=True)
class MemberParser(parser_base.Parser[disnake.Member]):
    r"""Asynchronous parser type with support for members.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.Member:
        """Load a member from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a member.

        Raises
        ------
        :class:`LookupError`:
            A member with the id stored in the ``argument`` could not be found.

        """
        member_id = await self.int_parser.loads(argument)

        maybe_member = di.resolve_dependency(disnake.Member, None)
        if maybe_member and maybe_member.id == member_id:
            return maybe_member

        guild = di.resolve_dependency(disnake.Guild)
        member = guild.get_member(member_id)
        if member:
            return member

        if self.allow_api_requests:
            with contextlib.suppress(disnake.HTTPException):
                return await guild.fetch_member(member_id)

        msg = f"Could not find a member with id {argument!r}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Member, /) -> str:
        """Dump a user into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)



"""Implementation of parser base classes upon which actual parsers are built."""

from __future__ import annotations

import typing

import attrs

from disnake_compass.api import parser as parser_api

if typing.TYPE_CHECKING:
    import typing_extensions

__all__: typing.Sequence[str] = (
    "Parser",
    "get_parser",
    "register_parser",
)

ParserT = typing.TypeVar("ParserT", bound=parser_api.Parser)

_PARSERS: dict[type, type[parser_api.Parser[typing.Any]]] = {}
_REV_PARSERS: dict[
    type[parser_api.Parser[typing.Any]],
    tuple[type, ...],
] = {}
_PARSER_PRIORITY: dict[type[parser_api.Parser[typing.Any]], int] = {}


def _issubclass(
    cls: type,
    class_or_tuple: type | tuple[type, ...],
) -> bool:
    try:
        return issubclass(cls, class_or_tuple)

    except TypeError:
        if isinstance(class_or_tuple, tuple):
            return any(cls is cls_ for cls_ in class_or_tuple)

        return cls is class_or_tuple


def register_parser(
    parser: type[parser_api.Parser[parser_api.ParserType]],
    *types: type[parser_api.ParserType],
    priority: int = 0,
    force: bool = True,
) -> None:
    """Register a parser class as the default parser for the provided type.

    The default parser will automatically be used for any field annotated
    with that type. For example, the default parser for integers is
    :class:`disnake_compass.IntParser`, an instance of which will automatically be
    assigned to any custom id fields annotated with `int`.

    Parameters
    ----------
    parser:
        The parser to register.
    *types:
        The types for which to register the provided parser as the default.
    priority:
        When a type has multiple parsers registered to it, priority is used to
        determine which parser to use.
    force:
        Whether or not to overwrite existing defaults. Defaults to ``True``.

    """
    # This allows e.g. is_default_for=(Tuple[Any, ...],) so pyright doesn't complain.
    # The stored type will then still be tuple, as intended.
    types = tuple(typing.get_origin(type_) or type_ for type_ in types)
    setter = dict.__setitem__ if force else dict.setdefault  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    setter(_REV_PARSERS, parser, types)
    setter(_PARSER_PRIORITY, parser, priority)
    for type_ in types:
        setter(_PARSERS, type_, parser)


def register_parser_for(
    *is_default_for: type[typing.Any],
    priority: int = 0,
) -> typing.Callable[[type[ParserT]], type[ParserT]]:
    def wrapper(cls: type[ParserT]) -> type[ParserT]:
        register_parser(cls, *is_default_for, priority=priority)
        return cls

    return wrapper


def _get_parser_type(
    type_: type[parser_api.ParserType],
) -> type[parser_api.Parser[parser_api.ParserType]]:
    # Fast lookup...
    if type_ in _PARSERS:
        return _PARSERS[type_]

    # Slow lookup for subclasses of existing types...
    best_entry = max(
        (entry for entry, parser_types in _REV_PARSERS.items() if _issubclass(type_, parser_types)),
        default=None,
        key=_PARSER_PRIORITY.__getitem__,
    )
    if best_entry is not None:
        return best_entry

    message = f"No parser available for type {type_.__name__!r}."
    raise TypeError(message)


# TODO: Maybe cache this?
def get_parser(  # noqa: D417
    type_: type[parser_api.ParserType],
) -> parser_api.Parser[parser_api.ParserType]:
    r"""Get the default parser for the provided type.

    Note that type annotations such as ``Union[int, str]`` are also valid.

    Parameters
    ----------
    type\_:
        The type for which to return the default parser.

    Returns
    -------
    :class:`Parser`\[``_T``]:
        The default parser for the provided type.

    Raises
    ------
    :class:`TypeError`:
        Could not create a parser for the provided type.

    """
    # TODO: Somehow allow more flexibility here. It would at the very least
    #       be neat to be able to pick between strictly sync/async parsers
    #       (mainly for the purpose of not making api requests); but perhaps
    #       allowing the user to pass a filter function could be cool?
    origin = typing.get_origin(type_)
    return _get_parser_type(origin or type_).default(type_)


@typing.runtime_checkable
@attrs.define(slots=True)
class Parser(
    parser_api.Parser[parser_api.ParserType],
    typing.Protocol[parser_api.ParserType],
):
    """Class that handles parsing of one custom id field to and from a desired type.

    A parser contains two main methods, :meth:`loads` and :meth:`dumps`.
    ``loads``, like :func:`json.loads` serves to turn a string value into
    a different type. Similarly, ``dumps`` serves to convert that type
    back into a string.
    """

    @classmethod
    def default(  # noqa: D102
        cls,
        target_type: type[parser_api.ParserType],  # noqa: ARG003
        /,
    ) -> typing_extensions.Self:
        # <<Docstring inherited from parser_api.Parser>>
        return cls()

    @classmethod
    def default_types(cls) -> tuple[type, ...]:
        """Return the types for which this parser type is the default implementation.

        Returns
        -------
        Sequence[type]:
            The types for which this parser type is the default implementation.

        """
        return _REV_PARSERS[cls]

    async def loads(self, argument: typing.Any, /) -> parser_api.ParserType:  # noqa: D102, ANN401
        # <<Docstring inherited from parser_api.Parser>>
        ...

    async def dumps(self, argument: parser_api.ParserType, /) -> str:  # noqa: D102
        # <<Docstring inherited from parser_api.Parser>>
        ...



"""Field implementations extending :func:`attrs.field`."""

from __future__ import annotations

import enum
import functools
import typing

import attrs
import typing_extensions

if typing.TYPE_CHECKING:
    from disnake_compass.api import parser as parser_api

__all__: typing.Sequence[str] = ("field",)


_T = typing_extensions.TypeVar("_T", default=typing.Any)


class FieldMetadata(enum.Enum):
    """Enum containing keys for field metadata."""

    PARSER = enum.auto()
    """Metadata key to store parser information."""
    FIELDTYPE = enum.auto()
    """Metadata key to store field type information. See :class:`FieldType`."""


class FieldType(enum.Flag):
    """Flag containing field metadata values for the field type.

    Note that a field can only ever be one of these types. This is a flag for
    the sole reason of facilitating unions in lookups using :func:`get_fields`.
    """

    INTERNAL = enum.auto()
    """Internal field that does not show up in the component's init signature."""
    CUSTOM_ID = enum.auto()
    """Field parsed into/from the component's custom id."""
    SELECT = enum.auto()
    """Field parsed from a select component's selected values."""
    MODAL = enum.auto()
    """Field parsed from a modal component's modal values."""

    @classmethod
    def ALL(cls) -> FieldType:  # noqa: N802
        """Meta-value for all field types.

        Mainly intended for use in :func:`get_fields`.
        """
        return _ALL_FIELD_TYPES


_ALL_FIELD_TYPES = functools.reduce(FieldType.__or__, FieldType)
_ALL_FIELD_TYPES._name_ = "ALL()"  # This makes it render nicer in docs.


def get_parser(
    field: attrs.Attribute[typing.Any],
) -> parser_api.Parser[typing.Any] | None:
    """Get the user-provided parser of the provided field.

    Parameters
    ----------
    field:
        The field for which to get the :class:`disnake_compass.api.Parser`.

    Returns
    -------
    :class:`disnake_compass.api.Parser`
        The user-provided parser of the provided field.
    :data:`None`
        The field's parser was automatically inferred.

    """
    return field.metadata.get(FieldMetadata.PARSER)


def get_field_type(
    field: attrs.Attribute[typing.Any],
    default: FieldType | None = None,
) -> FieldType:
    """Get the :class:`FieldType` of the field.

    Parameters
    ----------
    field:
        The field of which to get the field type.
    default:
        The default value to use if the field doesn't have a :class:`FieldType`
        set.

    Returns
    -------
    :class:`FieldType`
        The type of the provided field.

    Raises
    ------
    :class:`TypeError`
        The provided field does not have a field type set, and no default was
        provided. The most common cause of this is using
        :func:`attrs.field` instead of :func:`disnake_compass.field() <.field>` to
        define a field.

    """
    if FieldMetadata.FIELDTYPE not in field.metadata:
        if default:
            return default

        msg = (
            f"Field {field.name!r} does not contain the proper metadata to be"
            f" recognised by disnake-compass. Please only construct fields"
            f" through functions provided by disnake-compass."
            f" See the 'disnake_compass.fields' submodule for possibilities."
        )
        raise TypeError(msg)

    return field.metadata[FieldMetadata.FIELDTYPE]


def is_field_of_type(field: attrs.Attribute[typing.Any], kind: FieldType) -> bool:
    """Check whether or not a field is marked as the provided :class:`FieldType`.

    Parameters
    ----------
    field:
        The field to check.
    kind:
        The :class:`FieldType` to check for.

    Returns
    -------
    :class:`bool`
        Whether the provided field was of the provided :class:`FieldType`.

    """
    set_type = field.metadata.get(FieldMetadata.FIELDTYPE)
    return bool(set_type and set_type & kind)  # Check if not None, then check if match.


def get_fields(
    cls: type,
    /,
    *,
    kind: FieldType = _ALL_FIELD_TYPES,
) -> typing.Sequence[attrs.Attribute[typing.Any]]:
    r"""Get the attributes of an attrs class.

    This wraps :func:`attrs.fields` to be less strict typing-wise and has
    special handling for internal fields.

    Parameters
    ----------
    cls:
        The class of which to get the fields.
    kind:
        The kind(s) of fields to return. Can be any combination of
        :class:`FieldType`\s.

    """
    return [field for field in attrs.fields(cls) if is_field_of_type(field, kind)]


def field(
    default: _T | attrs.NothingType = attrs.NOTHING,
    *,
    parser: parser_api.Parser[_T] | None = None,
) -> _T:
    r"""Define a custom ID field for the component.

    The type annotation for this field is used to parse incoming custom ids.

    This is a wrapper around :func:`attrs.field`.

    .. note::
        In most cases, simply using a typehint will suffice to define a field.
        This function is generally only needed if you wish to supply a default
        value or custom parser.

    .. note::
        Fields created this way always have ``kw_only=True`` set.

    Parameters
    ----------
    default:
        The default value for this field. The type of the default should match
        that of the type annotation.
    parser:
        The parser to use for converting this field to and from a string.
        The type of the parser should match that of the type annotation.

    Returns
    -------
    :func:`Field <attrs.field>`\[``T``]
        A new field with the provided default and/or parser.

    """
    return attrs.field(
        default=typing.cast(_T, default),
        kw_only=True,
        metadata={
            FieldMetadata.FIELDTYPE: FieldType.CUSTOM_ID,
            FieldMetadata.PARSER: parser,
        },
    )


def internal(
    default: _T,
    *,
    frozen: bool = False,
) -> _T:
    r"""Declare a field as internal.

    This is used internally to differentiate component parameters from
    user-defined custom id parameters.

    This is a wrapper around :func:`attrs.field`.

    .. note::
        Fields created this way always have ``kw_only=True`` set.

    Parameters
    ----------
    default:
        The default value for this field. The type of the default should match
        that of the type annotation.
    frozen:
        Whether or not the field should be marked frozen. A frozen field cannot
        be modified after the class has been created.

    Returns
    -------
    :func:`Field <attrs.field>`\[``T``]
        A new field with the provided default and frozen status.

    """
    return attrs.field(
        default=default,
        on_setattr=attrs.setters.frozen if frozen else None,
        metadata={FieldMetadata.FIELDTYPE: FieldType.INTERNAL},
    )



"""Default implementation of button-based components."""

from __future__ import annotations

import typing

import disnake

from disnake_compass import fields
from disnake_compass.api import component as component_api
from disnake_compass.impl.component import base as component_base

__all__: typing.Sequence[str] = ("RichButton",)


_AnyEmoji: typing.TypeAlias = str | disnake.PartialEmoji | disnake.Emoji


# God knows why this is needed, but if I don't do this, classes inheriting from
# RichButton see e.g. `fields.internal(default=None)` in the init signature.
internal = fields.internal


@typing.runtime_checkable
class RichButton(
    component_api.RichButton,
    component_base.ComponentBase,
    typing.Protocol,
):
    """The default implementation of a disnake-compass button.

    This works similar to a dataclass, but with some extra things to take into
    account.

    First and foremost, there are class variables for :attr:`label`,
    :attr:`style`, :attr:`emoji` and :attr:`disabled`. These set the
    corresponding attributes on the button class when they are sent to discord,
    and are meant to be overwritten by the user.

    Next, fields can be defined similarly to dataclasses, by means of a name,
    a type annotation, and an optional :func:`components.field` to set the
    default or a custom parser.

    Classes created in this way have auto-generated slots and an auto-generated
    ``__init__``. The init-signature contains all the custom id fields as
    keyword-only arguments.
    """

    event: typing.ClassVar[str] = "on_button_click"

    label: str | None = fields.internal(default=None)
    style: disnake.ButtonStyle = fields.internal(default=disnake.ButtonStyle.secondary)
    emoji: _AnyEmoji | None = fields.internal(default=None)
    disabled: bool = fields.internal(default=False)

    async def as_ui_component(self) -> disnake.ui.Button[None]:  # noqa: D102
        # <<docstring inherited from component_api.RichButton>>

        if not self.manager:
            message = "Cannot serialise components without a manager."
            raise RuntimeError(message)

        return disnake.ui.Button(
            style=self.style,
            label=self.label,
            disabled=self.disabled,
            emoji=self.emoji,
            custom_id=await self.manager.make_custom_id(self),
        )



"""Default implementation of the component manager api."""

from __future__ import annotations

import contextlib
import contextvars
import logging
import sys
import typing
import weakref

import attrs
import disnake
import typing_extensions

from disnake_compass import fields
from disnake_compass.api import component as component_api
from disnake_compass.internal import di, omit

__all__: typing.Sequence[str] = ("ComponentManager", "check_manager", "get_manager")


_LOGGER = logging.getLogger(__name__)
_ROOT = sys.intern("root")
_COMPONENT_EVENT = sys.intern("on_message_interaction")
_MODAL_EVENT = sys.intern("on_modal_submit")

_COMPONENT_CTX: contextvars.ContextVar[tuple[component_api.RichComponent, str]] = (
    contextvars.ContextVar("_COMPONENT_CTX")
)


T = typing.TypeVar("T")


class DependencyProviderFunc(typing.Protocol):
    def __call__(
        self,
        manager: ComponentManager,
        *dependencies: object,
    ) -> typing.AsyncGenerator[None, None]: ...


class DependencyProvider(typing.Protocol):
    def __call__(
        self,
        manager: ComponentManager,
        *dependencies: object,
    ) -> typing.AsyncContextManager[None]: ...


DependencyProviderFuncT = typing.TypeVar("DependencyProviderFuncT", bound=DependencyProviderFunc)


CallbackWrapperFunc: typing.TypeAlias = typing.Callable[
    ["ComponentManager", component_api.RichComponent, disnake.Interaction[disnake.Client]],
    typing.AsyncGenerator[None, None],
]
CallbackWrapper: typing.TypeAlias = typing.Callable[
    ["ComponentManager", component_api.RichComponent, disnake.Interaction[disnake.Client]],
    typing.AsyncContextManager[None],
]
CallbackWrapperFuncT = typing.TypeVar("CallbackWrapperFuncT", bound=CallbackWrapperFunc)


ExceptionHandlerFunc: typing.TypeAlias = typing.Callable[
    [
        "ComponentManager",
        component_api.RichComponent,
        disnake.Interaction[disnake.Client],
        Exception,
    ],
    typing.Coroutine[typing.Any, typing.Any, bool | None],
]
ExceptionHandlerFuncT = typing.TypeVar(
    "ExceptionHandlerFuncT",
    bound=ExceptionHandlerFunc,
)

RichComponentT = typing.TypeVar("RichComponentT", bound=component_api.RichComponent)
RichComponentType: typing.TypeAlias = type[component_api.RichComponent]

MessageComponents: typing.TypeAlias = (
    component_api.RichButton
    | disnake.ui.Button[typing.Any]
    | component_api.RichSelect
    | disnake.ui.StringSelect[typing.Any]
    | disnake.ui.ChannelSelect[typing.Any]
    | disnake.ui.RoleSelect[typing.Any]
    | disnake.ui.UserSelect[typing.Any]
    | disnake.ui.MentionableSelect[typing.Any]
)


def _to_ui_component(
    component: disnake.Button | disnake.BaseSelectMenu,
) -> disnake.ui.MessageUIComponent:
    if isinstance(component, disnake.Button):
        return disnake.ui.Button[None].from_component(component)
    if isinstance(component, disnake.StringSelectMenu):
        return disnake.ui.StringSelect[None].from_component(component)
    if isinstance(component, disnake.UserSelectMenu):
        return disnake.ui.UserSelect[None].from_component(component)
    if isinstance(component, disnake.RoleSelectMenu):
        return disnake.ui.RoleSelect[None].from_component(component)
    if isinstance(component, disnake.MentionableSelectMenu):
        return disnake.ui.MentionableSelect[None].from_component(component)
    if isinstance(component, disnake.ChannelSelectMenu):
        return disnake.ui.ChannelSelect[None].from_component(component)

    msg = f"Expected a message component type, got {type(component).__name__!r}."
    raise TypeError(msg)


_MAX_COUNT = 1 << 8 - 1  # 1 byte, starting at 0


def _minimise_count(count: int) -> str:
    # We only need to support counts up to 25, as that is the
    # maximum number of components that can go on a message.
    # Byte-length 1 should support a range of 0~255 inclusive.
    byte = count.to_bytes(1, "little")
    # Decode into a charset that supports these bytes as a single char.
    return byte.decode("latin-1")


_COUNT_CHARS: typing.Final[tuple[str, ...]] = tuple(
    map(_minimise_count, range(25)),
)
_DEFAULT_SEP: typing.Final[str] = sys.intern("|")
_DEFAULT_COUNT: typing.Final = True


@contextlib.asynccontextmanager
async def default_dependency_provider(
    manager: component_api.ComponentManager,  # noqa: ARG001
    *dependencies: object,
) -> typing.AsyncGenerator[None, None]:
    tokens = di.register_dependencies(
        *(dependency for dependency in dependencies if dependency is not None),
    )

    yield

    di.reset_dependencies(tokens)


@contextlib.asynccontextmanager
async def default_callback_wrapper(
    manager: component_api.ComponentManager,  # noqa: ARG001
    component: component_api.RichComponent,  # noqa: ARG001
    interaction: disnake.Interaction[disnake.Client],  # noqa: ARG001
) -> typing.AsyncGenerator[None, None]:
    """Wrap a callback for a component manager.

    This is the default implementation, and is effectively a no-op.
    """
    yield


async def default_exception_handler(
    manager: component_api.ComponentManager,
    component: component_api.RichComponent,
    interaction: disnake.Interaction[disnake.Client],  # noqa: ARG001
    exception: Exception,
) -> bool:
    """Handle an exception that occurs during execution of a component callback.

    This is the default implementation, and simply passes the exception down.
    If it is passed down to the root logger, and the root logger also has this
    default implementation, the exception is logged.
    """
    if manager.name is not _ROOT:
        # Not the root manager, try passing down.
        return False

    # We're at the root logger, and the exception remains unhandled. Log it.

    exc_info = (
        type(exception),
        exception,
        exception.__traceback__.tb_next if exception.__traceback__ else None,
    )

    _LOGGER.exception(
        "An exception was caught on manager %r while handling the callback of"
        " component %r, registered to manager %r:",
        manager.name,
        component,
        component.manager.name if component.manager else "<unknown>",
        exc_info=exc_info,
    )

    return True


@attrs.define
class _ModuleData:
    name: str
    id: int

    @classmethod
    def from_object(cls, obj: object) -> typing_extensions.Self:
        module = sys.modules[obj.__module__]
        return cls(obj.__module__, id(module))

    def is_active(self) -> bool:
        if self.name not in sys.modules:
            return False

        return self.id == id(sys.modules[self.name])

    def is_reload_of(self, other: typing_extensions.Self) -> bool:
        return self.name == other.name and self.id != other.id


class ComponentManager(component_api.ComponentManager):
    """The standard implementation of a component manager.

    Component managers keep track of disnake-compass' special components
    and ensure they smoothly communicate with disnake's clients.

    To register a component to a component manager, use :meth:`register`.
    Without registering your components, they will remain unresponsive.

    To get an instance of a component manager, use :func:`get_manager`. This
    will automatically create missing managers if needed, much like
    :func:`logging.getLogger`. Similarly, managers feature a parent-child
    hierarchy in the same way loggers do. For example, a manager named
    "foo.bar" would be a child of the manager named "foo".

    The topmost manager will always be the root manager, which can be acquired
    through calling :func:`get_manager` without passing a name.

    When a component is invoked on - for example - a manager "foo.bar", it will
    wrap the callback in the :meth:`as_callback_wrapper` wrappers bubbling up.
    That is, the callback is wrapped by the root manager, then "foo",
    then "foo.bar", and finally invoke the callback.

    If any exceptions occur during the wrapping or invocation of the callback,
    the managers' exception handlers will be invoked starting from "foo.bar",
    then "foo", and finally the root manager. If any exception handler returns
    ``True``, the exception is considered handled and any remaining exception
    handlers are skipped.

    Parameters
    ----------
    name:
        The name of the component manager. This should be unique for all live
        component managers.
    count:
        Whether the component manager should insert *one* count character to
        resolve duplicates. Normally, sending two components with the same
        custom id would error. Enabling this ensures custom ids are unique
        by appending an incrementing character. This costs 1 character,
        effectively reducing the maximum custom id length to 99 characters.

        If not set, the manager will use its parents' settings. The default
        set on the root manager is ``True``.
    sep:
        The character(s) to use as separator between custom id parts.

        If not set, the manager will use its parents' settings. The default
        set on the root manager is ``"|"``.
    client:
        The client to which to register this manager. This can be specified at any
        point through :meth:`.add_to_client`.

    """

    __slots__: typing.Sequence[str] = (
        "_children",
        "_client",
        "_components",
        "_count",
        "_counter",
        "_identifiers",
        "_module_data",
        "_name",
        "_sep",
        "handle_exception",
        "set_invocation_dependencies",
        "wrap_callback",
    )

    _client: disnake.Client | None
    _children: set[ComponentManager]
    _components: weakref.WeakValueDictionary[str, RichComponentType]
    _count: bool | None
    _counter: int
    _identifiers: dict[str, str]
    _module_data: dict[str, _ModuleData]
    _name: str
    _sep: str | None

    def __init__(
        self,
        name: str,
        *,
        count: bool | None = None,
        sep: str | None = None,
        client: disnake.Client | None = None,
    ) -> None:
        self._name = name
        self._children = set()
        self._components = weakref.WeakValueDictionary()
        self._identifiers = {}
        self._count = count
        self._counter = 0
        self._module_data = {}
        self._sep = sep
        self.set_invocation_dependencies: DependencyProvider = default_dependency_provider
        self.wrap_callback: CallbackWrapper = default_callback_wrapper
        self.handle_exception: ExceptionHandlerFunc = default_exception_handler

        if client:
            self.add_to_client(client)

    def __repr__(self) -> str:
        return f"ComponentManager(name={self.name})"

    @property
    @typing_extensions.deprecated("Please use client instead.")
    def bot(self) -> disnake.Client:
        """An alias of :attr:`client`."""
        return self.client

    @property
    def client(self) -> disnake.Client:
        """The client to which this manager is registered.

        If the manager has not yet been registered, this raises an exception.

        .. note::
            This is recursively accessed for all the parents of this manager.
            For example, if ``get_manager().client`` is set, then any of its
            children ``get_manager("foo.bar").client`` will also return that
            same client instance.

            It is therefore generally recommended to set the client on the root
            manager so that all other managers automatically have access to it.
        """
        client = _recurse_parents_getattr(self, "_client", None)
        if client:
            return client

        msg = f"Component manager {self.name!r} is not yet registered to a client."
        raise RuntimeError(msg)

    @property
    def name(self) -> str:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        return self._name

    @property
    def children(self) -> set[ComponentManager]:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        return self._children

    @property
    def components(self) -> typing.Mapping[str, RichComponentType]:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        return self._components

    @property
    def count(self) -> bool:
        """Whether or not this manager should add a count character to custom ids.

        This prevents an error when two components with otherwise equal custom
        ids are sent.

        By default, this is set to :obj:`True`. This can be changed using
        :meth:`config`.

        .. note::
            This is recursively checked for all the parents of this manager.
            For example, if ``get_manager("foo").count == True``, then its
            child ``get_manager("foo.bar").count`` will also return ``True``
            unless explicitly set to ``False``.

        .. warning::
            As this takes 1 character, the effective maximum custom id length
            is reduced to 99 characters.
        """
        return _recurse_parents_getattr(self, "_count", _DEFAULT_COUNT)

    @property
    def counter(self) -> int:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        return self._counter

    @property
    def sep(self) -> str:
        """The separator used to delimit parts of the custom ids of this manager.

        By default, this is set to "|". This can be changed using
        :meth:`config`.

        .. note::
            This is recursively accessed for all the parents of this manager.
            For example, if ``get_manager("foo").sep == "|"``, then its
            child ``get_manager("foo.bar").sep`` will also return ``"|"``
            unless explicitly set to some other value.
        """
        return _recurse_parents_getattr(self, "_sep", _DEFAULT_SEP)

    @property
    def parent(self) -> ComponentManager | None:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        if "." not in self.name:
            # Return the root manager if this is not the root manager already.
            return None if self.name is _ROOT else get_manager(_ROOT)

        root, _ = self.name.rsplit(".", 1)
        return get_manager(root)

    def config(
        self,
        count: omit.OmittedNoneOr[bool] = omit.Omitted,
        sep: omit.OmittedNoneOr[str] = omit.Omitted,
    ) -> None:
        """Set configuration options on this manager."""
        if not omit.is_omitted(count):
            self._count = count

        if not omit.is_omitted(sep):
            self._sep = sep

    def make_identifier(self, component_type: RichComponentType) -> str:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        return component_type.__name__

    def get_identifier(  # noqa: D102
        self,
        custom_id: str,
    ) -> tuple[str, typing.Sequence[str]]:
        # <<docstring inherited from api.components.ComponentManager>>

        name, *params = custom_id.split(self.sep)

        if self.count and name.endswith(_COUNT_CHARS):
            # Count is always the single last character in the name part.
            return name[:-1], params

        return name, params

    def increment(self) -> str:  # noqa: D102
        count = _minimise_count(self._counter)

        self._counter += 1
        if self._counter > _MAX_COUNT:
            self._counter = 0

        return count

    async def make_custom_id(  # noqa: D102
        self,
        component: component_api.RichComponent,
    ) -> str:
        # <<docstring inherited from api.components.ComponentManager>>

        identifier = self._identifiers[type(component).__name__]

        if self.count:
            identifier = identifier + self.increment()

        dumped_params = await component.factory.dump_params(component)

        return self.sep.join([identifier, *dumped_params.values()])

    async def parse_message_interaction(  # noqa: D102
        self,
        interaction: disnake.Interaction[disnake.Client],
    ) -> component_api.RichComponent | None:
        # <<docstring inherited from api.components.ComponentManager>>
        if isinstance(interaction, disnake.MessageInteraction):
            return await self.parse_raw_component(interaction.component)

        raise NotImplementedError

    async def parse_raw_component(
        self,
        component: disnake.Button | disnake.BaseSelectMenu,
    ) -> component_api.RichComponent | None:
        """Parse a rich message component from a disnake raw component.

        .. note::
            This method only works for components registered to this manager.

        Parameters
        ----------
        component:
            The raw message component that is to be turned into a rich
            component.

        Returns
        -------
        :class:`RichComponent`
            The newly created component.
        :obj:`None`:
            The provided component could not be parsed into a rich component
            that is registered to this manager.

        """
        custom_id = component.custom_id
        if not custom_id:
            return None

        identifier, params = self.get_identifier(custom_id)

        if identifier not in self._components:
            return None

        component_type = self._components[identifier]
        module_data = self._module_data[identifier]

        if not module_data.is_active():
            # NOTE: This occurs if:
            #       - The module on which the component is defined was unloaded.
            #       - The module on which the component is defined was reloaded
            #         and the component was never overwritten. It could either
            #         have been removed, or simply no longer be registered. The
            #         component *should* therefore be unresponsive.
            #
            #       Since we do not want to fire components that (to the user)
            #       do not exist anymore, we should remove them from the
            #       manager and return None.
            self.deregister_component(component_type)
            return None

        component_params = {
            field.name: getattr(component, field.name)
            for field in fields.get_fields(
                component_type,
                kind=fields.FieldType.INTERNAL,
            )
        }

        return await component_type.factory.build_component(
            params,
            component_params=component_params,
        )

    async def parse_message_components(
        self,
        message: disnake.Message,
    ) -> tuple[
        typing.Sequence[typing.Sequence[MessageComponents]],
        typing.Sequence[component_api.RichComponent],
    ]:
        """Parse all components on a message into rich components or ui components.

        This method is particularly useful if you wish to modify multiple
        components attached to a given message.

        This returns a structure of components that can be directly passed into
        any send method's component parameters, and a separate sequence
        containing all rich components for easier editing.

        Parameters
        ----------
        message:
            The message of which to parse all components.

        Returns
        -------
        :class:`tuple`[:class:`Sequence`[:class:`Sequence`[:obj:`MessageComponents`]], :class:`Sequence`[:class:`RichComponent`]]
            A tuple containing:

            - A nested structure of sequences that contains the parsed message
            components. The outer sequence can contain a maximum of 5 inner
            sequences, which can each contain up to five components; as per
            Discord API spec.

            - A sequence containing only the rich components to facilitate
            easier modification of the components.

            These objects share the same component instances, so any changes
            made to components inside the separate sequence will also reflect
            on the nested structure.

        """  # noqa: E501
        new_rows: list[list[MessageComponents]] = []
        rich_components: list[component_api.RichComponent] = []

        current_component, current_component_id = _COMPONENT_CTX.get((None, None))
        should_test = current_component is not None

        for row in message.components:
            new_row: list[MessageComponents] = []
            new_rows.append(new_row)

            for component in row.children:
                if should_test and component.custom_id == current_component_id:
                    should_test = False
                    new_component = current_component

                else:
                    new_component = await self.parse_raw_component(component)

                if new_component:
                    rich_components.append(new_component)
                    assert isinstance(
                        new_component,
                        component_api.RichButton | component_api.RichSelect,
                    )

                else:
                    new_component = _to_ui_component(component)

                new_row.append(new_component)

        return new_rows, rich_components

    async def finalise_components(
        self,
        components: typing.Sequence[typing.Sequence[MessageComponents]],
    ) -> disnake.ui.Components[disnake.ui.MessageUIComponent]:
        """Finalise the output of :meth:`parse_message_components` back into disnake ui components.

        Parameters
        ----------
        components:
            A sequence of rows of components, which can be any combination of
            disnake ui components and rich components. The rich components are
            automatically cast to their equivalent disnake ui components so
            that they can be sent as an interaction response.

        Returns
        -------
        disnake.ui.Components[disnake.ui.MessageUIComponent]:
            A disnake-compatible structure of sendable components.

        """
        finalised: list[list[disnake.ui.MessageUIComponent]] = []

        for row in components:
            new_row: list[disnake.ui.MessageUIComponent] = []
            finalised.append(new_row)

            for component in row:
                if isinstance(component, component_api.RichButton | component_api.RichSelect):
                    new_row.append(await component.as_ui_component())  # pyright: ignore[reportArgumentType]
                else:
                    new_row.append(component)

        return finalised

    # Identifier and component: function call, return component
    @typing.overload
    def register(
        self,
        component_type: type[RichComponentT],
        *,
        identifier: str | None = None,
    ) -> type[RichComponentT]: ...

    # Only identifier: nested decorator, return callable that registers and
    # returns the component.
    @typing.overload
    def register(
        self,
        *,
        identifier: str | None = None,
    ) -> typing.Callable[[type[RichComponentT]], type[RichComponentT]]: ...

    def register(
        self,
        component_type: type[RichComponentT] | None = None,
        *,
        identifier: str | None = None,
    ) -> type[RichComponentT] | typing.Callable[[type[RichComponentT]], type[RichComponentT]]:
        """Register a component to this component manager.

        This is the decorator interface to :meth:`register_component`.
        """
        if component_type is not None:
            return self.register_component(component_type, identifier=identifier)

        def wrapper(
            component_type: type[RichComponentT],
        ) -> type[RichComponentT]:
            return self.register_component(component_type, identifier=identifier)

        return wrapper

    def register_component(  # noqa: D102
        self,
        component_type: type[RichComponentT],
        *,
        identifier: str | None = None,
    ) -> type[RichComponentT]:
        # <<docstring inherited from api.components.ComponentManager>>
        resolved_identifier = identifier or self.make_identifier(component_type)
        module_data = _ModuleData.from_object(component_type)

        root_manager = get_manager(_ROOT)

        if resolved_identifier in root_manager._components:  # noqa: SLF001
            # NOTE: This occurs when a component is registered while another
            #       component with the same identifier already exists.
            #
            #       We now have two options:
            #       - This is caused by a reload. In this case, we expect the
            #         module name to remain unchanged and the module id to have
            #         changed. We can safely overwrite the old component.
            #       - This is an actual user error. If we were to silently
            #         overwrite the old component, it would unexpectedly go
            #         unresponsive. Instead, we raise an exception to the user.
            old_module_data = root_manager._module_data[resolved_identifier]  # noqa: SLF001
            if not module_data.is_reload_of(old_module_data):
                message = (
                    "Cannot register component with duplicate identifier"
                    f" {identifier!r}. (Original defined in module"
                    f" {old_module_data.name!r}, duplicate defined in module"
                    f" {module_data.name!r})"
                )
                raise RuntimeError(message)

        # Register to current manager and all parent managers.
        component_type.manager = self

        for manager in _recurse_parents(self):
            manager._components[resolved_identifier] = component_type  # noqa: SLF001
            manager._identifiers[component_type.__name__] = resolved_identifier  # noqa: SLF001
            manager._module_data[resolved_identifier] = module_data  # noqa: SLF001

        return component_type

    def deregister_component(  # noqa: D102
        self,
        component_type: RichComponentType,
    ) -> None:
        # <<docstring inherited from api.components.ComponentManager>>

        identifier = self.make_identifier(component_type)
        component = self._components[identifier]

        if not component.manager:
            message = (
                f"Component {component_type.__name__!r} is not registered to a component manager."
            )
            raise TypeError(message)

        if not isinstance(component.manager, ComponentManager):
            # This should honestly never happen unless the user does some
            # really weird stuff.
            # TODO: Maybe think of an error message for this.
            raise TypeError

        # Deregister from the current manager and all parent managers.
        for manager in _recurse_parents(component.manager):
            manager._components.pop(identifier)  # noqa: SLF001
            manager._module_data.pop(identifier)  # noqa: SLF001

    def add_to_client(self, client: disnake.Client, /) -> None:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        # Ensure we don't duplicate the listeners.
        if (
            self.invoke_component in client.extra_events.get(_COMPONENT_EVENT, [])
            # or self.invoke_component in client.extra_events.get(_MODAL_EVENT, [])
        ):  # fmt: skip
            message = "This component manager is already registered to this client."
            raise RuntimeError(message)

        client.add_listener(self.invoke_component, _COMPONENT_EVENT)
        # client.add_listener(self.invoke, _MODAL_EVENT)  # noqa: ERA001

        self._client = client

    @typing_extensions.deprecated("Please use add_to_client() instead.")
    def add_to_bot(self, bot: disnake.Client) -> None:  # noqa: D102
        self.add_to_client(bot)

    def remove_from_client(self, client: disnake.Client, /) -> None:  # noqa: D102
        # <<docstring inherited from api.components.ComponentManager>>

        # client.remove_listener silently ignores if the event doesn't exist,
        # so we manually handle raising an exception for it.
        if not (
            self.invoke_component in client.extra_events.get(_COMPONENT_EVENT, [])  # noqa: E713
            # and self.invoke_component in client.extra_events.get(_MODAL_EVENT, [])
        ):  # fmt: skip
            message = "This component manager is not yet registered to this client."
            raise RuntimeError(message)

        client.remove_listener(self.invoke_component, _COMPONENT_EVENT)
        # client.remove_listener(self.invoke_component, _MODAL_EVENT)  # noqa: ERA001

    @typing_extensions.deprecated("Please use remove_from_client() instead.")
    def remove_from_bot(self, bot: disnake.Client) -> None:  # noqa: D102
        self.remove_from_client(bot)

    def as_dependency_provider(self, func: DependencyProviderFuncT) -> DependencyProviderFuncT:
        """Register a callback as this manager's dependency provider.

        By default, this registers everything passed as a dependency.

        A dependency provider MUST be an async function with ONE yield statement.
        - Any code before the yield statement is run before the interaction is
        parsed,
        - The interaction is parsed at the yield statement,
        - Any code after the yield statement is run after the interaction has
        been parsed. This can be used for cleanup.

        It is therefore also possible to use context managers over the yield
        statement, to automatically handle resource management.

        As this runs *before* we know which manager a component belongs to (if
        at all!), this is run only for the manager(s) registered to a client
        with :meth:`add_to_client`.

        Examples
        --------
        .. code-block:: python

            manager = get_manager()


            @manager.as_dependency_provider
            async def provider(manager, *dependencies):
                tokens = di.register_dependencies(*dependencies)
                yield
                di.reset_dependencies(tokens)

        Parameters
        ----------
        func:
            The callback to register. This must be an async function that takes
            the component manager as the first argument, and any number of
            dependencies after. The function must have a single ``yield``
            statement that yields ``None``.

        Returns
        -------
        Callable[[:class:`ComponentManager`, ...], AsyncGenerator[None, None]]
            The function that was just registered.

        """
        self.set_invocation_dependencies = contextlib.asynccontextmanager(func)
        return func

    def as_callback_wrapper(self, func: CallbackWrapperFuncT) -> CallbackWrapperFuncT:
        """Register a callback as this managers' callback wrapper.

        By default, this is essentially a no-op.

        A callback wrapper MUST be an async function with ONE yield statement.
        - Any code before the yield statement is run before the component
        callback is invoked,
        - The component is invoked at the yield statement,
        - Any code after the yield statement is run after the component
        callback is invoked. This can be used for cleanup.

        It is therefore also possible to use context managers over the yield
        statement, to automatically handle resource management.

        In case this manager has a parent manager, the parent's callback
        wrapper will be used first, starting all the way at the root manager.
        For example, on a manager named "foo.bar", the callback will first be
        wrapped by the root manager, then by "foo", then by "foo.bar", and only
        then will the component callback finally be invoked.

        Note that any exceptions raised in any callback wrapper will cancel any
        other active callback wrappers and propagate the exception to the
        manager's error handler.

        Examples
        --------
        .. code-block:: python

            manager = get_manager()


            @manager.as_callback_wrapper
            async def wrapper(component, interaction):
                print(f"User {inter.author.name} invoked {type(component).__name__}.)
                yield
                print(f"Successfully ran callback for {type(component).__name__}.)

        Parameters
        ----------
        func:
            The callback to register. This must be an async function that takes
            the component manager as the first argument, the component as the
            second argument, and the interaction as the last. The function must
            have a single ``yield``-statement that yields ``None``.

        Returns
        -------
        Callable[[:class:`RichComponent`, :class:`disnake.Interaction[disnake.Client]`], AsyncGenerator[None, None]]
            The function that was just registered.

        """  # noqa: E501
        self.wrap_callback = contextlib.asynccontextmanager(func)
        return func

    def as_exception_handler(
        self,
        func: ExceptionHandlerFuncT,
    ) -> ExceptionHandlerFuncT:
        """Register a callback as this managers' error handler.

        By default, this simply logs the exception and keeps it from
        propagating.

        An error handler should return a boolean or ``None``:
        - ``True`` if the error was successfully handled and should not be
        propagated further.
        - ``False`` or ``None`` if the error was *not* successfully handled and
        should be passed to the next error handler in line.

        Note that it is therefore also possible to use context managers over
        the yield statement.

        In case this manager has a parent manager, the parent's error handler
        will be used if this one returns ``False`` or ``None``. For example,
        of a manager named "foo.bar", any exceptions will first be handled by
        "foo.bar", if that fails it will be handled by "foo", and finally if
        that also fails it will be handled by the root handler.

        Examples
        --------
        .. code-block:: python

            manager = get_manager()


            @manager.as_exception_handler
            async def wrapper(component, interaction, exception):
                if isinstance(exception, TypeError):
                    return True  # Silently ignore any TypeErrors

                return False  # Propagate all other errors.

        Parameters
        ----------
        func:
            The callback to register. This must be an async function that takes
            the component manager as the first argument, the component as the
            second, the interaction as the third, and the exception as the last.
            The function must return ``True`` to indicate that the error was
            handled successfully, or either ``False`` or ``None`` to indicate
            the opposite.

        Returns
        -------
        Callable[[:class:`RichComponent`, :class:`disnake.Interaction[disnake.Client]`, :class:`Exception`], None]
            The function that was just registered.

        """  # noqa: E501
        self.handle_exception = func
        return func

    async def _invoke_component(
        self,
        interaction: disnake.MessageInteraction[disnake.Client],
        /,
    ) -> None:
        # First, we check if the component is managed.
        component = await self.parse_message_interaction(interaction)
        if not (component and component.manager):
            # If the component was found, the manager is guaranteed to be
            # defined but we need the extra check for type-safety.
            return

        if not isinstance(component.manager, ComponentManager):
            # This should honestly never happen unless the user does some
            # really weird stuff.
            # TODO: Maybe think of an error message for this.
            raise TypeError

        # We traverse the managers in reverse: root first, then child, etc.
        # until we reach the component's actual manager. Therefore, we first
        # store all managers in a list, so that we can call reversed() on it
        # later.
        # This applies only to the callback wrappers. Error handlers are called
        # starting from the actual manager and propagated down to the root
        # manager if the error was left unhandled.
        managers = list(_recurse_parents(component.manager))

        assert interaction.component.custom_id
        ctx_value = (component, interaction.component.custom_id)
        component_ctx_token = _COMPONENT_CTX.set(ctx_value)

        try:
            async with contextlib.AsyncExitStack() as stack:
                # Enter all the context managers...
                for manager in reversed(managers):
                    await stack.enter_async_context(
                        manager.wrap_callback(manager, component, interaction),
                    )

                # If none raised, we run the callback.
                await component.callback(interaction)

        except Exception as exception:  # noqa: BLE001
            # Blanket exception catching is desired here as it's meant to
            # redirect all non-system errors to the error handler.

            for manager in managers:
                if await manager.handle_exception(
                    manager,
                    component,
                    interaction,
                    exception,
                ):
                    # If an error handler returns True, consider the error
                    # handled and skip the remaining handlers.
                    break

        finally:
            _COMPONENT_CTX.reset(component_ctx_token)

    async def invoke_component(  # noqa: D102
        self,
        interaction: disnake.MessageInteraction[disnake.Client],
        /,
        *,
        with_di: bool = True,
    ) -> None:
        # <<docstring inherited from api.components.ComponentManager>>

        if with_di:
            async with self.set_invocation_dependencies(
                self,
                interaction,
                interaction.guild,
                interaction.bot,
                interaction.channel,
                interaction.author,
                # XXX:  Potential edge-case here where a parser needs a user
                #       but we can only provide a member.
            ):
                await self._invoke_component(interaction)

        else:
            await self._invoke_component(interaction)

    def make_button(  # noqa: PLR0913
        self,
        identifier: str,
        *,
        as_root: bool = True,
        label: omit.Omissible[str | None] = omit.Omitted,
        style: omit.Omissible[disnake.ButtonStyle] = omit.Omitted,
        emoji: omit.Omissible[component_api.AnyEmoji | None] = omit.Omitted,
        disabled: omit.Omissible[bool] = omit.Omitted,
        **kwargs: object,
    ) -> component_api.RichButton:
        """Make an instance of the button class with the provided identifier.

        Parameters
        ----------
        as_root:
            Whether to use the root manager to get the component. This defaults
            to ``True`` so that any externally registered button can be built.
        identifier:
            The identifier of the button that is to be instantiated.
        label:
            The label to use. If not provided, uses the button class' default.
        style: disnake.ButtonStyle
            The style to use. If not provided, uses the button class' default.
        emoji:
            The emoji to use. If not provided, uses the button class' default.
        disabled:
            Whether or not to disable the button. If not provided, uses the
            button class' default.
        **kwargs:
            Any remaining keyword arguments are passed to the button's ``__init__``.

        Returns
        -------
        :class:`disnake_compass.api.RichButton`
            The newly created button.

        Raises
        ------
        :class:`KeyError`
            The provided identifier does not belong to a registered component.
        :class:`TypeError`
            The provided identifier belongs to a component that is not a button.
        :class:`Exception`
            Any exception raised during button instantiation is propagated as-is.

        """
        if label is not omit.Omitted:
            kwargs["label"] = label
        if style is not omit.Omitted:
            kwargs["style"] = style
        if emoji is not omit.Omitted:
            kwargs["emoji"] = emoji
        if disabled is not omit.Omitted:
            kwargs["disabled"] = disabled

        manager = get_manager(_ROOT) if as_root else self
        component_type = manager.components[identifier]
        component = component_type(**kwargs)

        # NOTE: We sadly cannot use issubclass-- maybe make a custom issubclass
        #       implementation that works with protocols with non-method members
        #       given a couple assumptions.
        if isinstance(component, component_api.RichButton):
            return component

        msg = (
            f"Expected identifier {identifier!r} to point to a button class,"
            f" got {component_type.__name__}."
        )
        raise TypeError(msg)

    def make_select(  # noqa: PLR0913
        self,
        identifier: str,
        *,
        as_root: bool = True,
        placeholder: omit.Omissible[str | None] = omit.Omitted,
        min_values: omit.Omissible[int] = omit.Omitted,
        max_values: omit.Omissible[int] = omit.Omitted,
        disabled: omit.Omissible[bool] = omit.Omitted,
        options: omit.Omissible[list[disnake.SelectOption]] = omit.Omitted,
        **kwargs: object,
    ) -> component_api.RichSelect:
        """Make an instance of the string select class with the provided identifier.

        Parameters
        ----------
        as_root:
            Whether to use the root manager to get the component. This defaults
            to ``True`` so that any externally registered select can be built.
        identifier:
            The identifier of the button that is to be instantiated.
        placeholder:
            The placeholder to use. If not provided, uses the select class' default.
        min_values:
            The minimum number of values a user is allowed to select. If not
            provided, uses the select class' default.
        max_values:
            The maximum number of values a user is allowed to select. If not
            provided, uses the select class' default.
        disabled:
            Whether or not to disable the button. If not provided, uses the
            select class' default.
        options:
            The options to use. If not provided, uses the select class' default.
        **kwargs:
            Any remaining keyword arguments are passed to the select's ``__init__``.

        Returns
        -------
        :class:`disnake_compass.api.RichStringSelect`
            The newly created string select.

        Raises
        ------
        :class:`KeyError`
            The provided identifier does not belong to a registered component.
        :class:`TypeError`
            The provided identifier belongs to a component that is not a string select.
        :class:`Exception`
            Any exception raised during button instantiation is propagated as-is.

        """
        # NOTE: This currently only supports StringSelects

        if placeholder is not omit.Omitted:
            kwargs["placeholder"] = placeholder
        if min_values is not omit.Omitted:
            kwargs["min_values"] = min_values
        if max_values is not omit.Omitted:
            kwargs["max_values"] = max_values
        if disabled is not omit.Omitted:
            kwargs["disabled"] = disabled
        if options is not omit.Omitted:
            kwargs["options"] = options

        manager = get_manager(_ROOT) if as_root else self
        component_type = manager.components[identifier]
        component = component_type(**kwargs)

        # NOTE: We sadly cannot use issubclass-- maybe make a custom issubclass
        #       implementation that works with protocols with non-method members
        #       given a couple assumptions.
        if isinstance(component, component_api.RichSelect):
            return component

        msg = (
            f"Expected identifier {identifier!r} to point to a select class,"
            f" got {component_type.__name__}."
        )
        raise TypeError(msg)


_MANAGER_STORE: typing.Final[dict[str, ComponentManager]] = {}


def _recurse_parents(manager: ComponentManager) -> typing.Iterator[ComponentManager]:
    yield manager
    while manager := manager.parent:  # pyright: ignore[reportAssignmentType]
        yield manager


def _recurse_parents_getattr(
    manager: ComponentManager,
    attribute: str,
    default: T,
) -> T:
    for parent in _recurse_parents(manager):
        value = getattr(parent, attribute)
        if value is not None:
            return value

    return default


def get_manager(name: str | None = None) -> ComponentManager:
    """Get a manager by name, or create one if it does not yet exist.

    Calling :func:`get_manager` without specifying a name returns the root
    manager. The root manager is -- unless explicitly modified by the user --
    guaranteed to be the lowest-level manager, with no parents.

    Managers follow a parent-child hierarchy. For example, a manager "foo.bar"
    would be a child of "foo". Any components registered to "foo.bar" would
    also be accessible to manager "foo". This means that the root manager
    has access to all components.

    To register a component to a manager, use :meth:`ComponentManager.register`.
    To ensure component callbacks are invoked, the manager must first be linked
    to a client. This is done using :meth:`ComponentManager.add_to_client`.
    Since parents have access to the components of their children, it is often
    sufficient to bind only the root manager to a client.

    It is generally recommended to use a separate manager per extension, though
    you can share the same manager between files by using the same name, if
    desired.

    Further configuration of managers can be done through
    :meth:`ComponentManager.config`.

    Parameters
    ----------
    name: str
        The name of the component. If not provided, the root manager is
        returned.

    Returns
    -------
    :class:`ComponentManager`:
        A component manager with the desired name. If a component manager with
        this name already existed before calling this function, that same
        manager is returned. Otherwise, a new manager is created.

    """
    if name is None:
        # TODO: Maybe use a sentinel:
        #       - auto-infer name if sentinel,
        #       - return root logger if None was passed explicitly.
        name = _ROOT

    if name in _MANAGER_STORE:
        return _MANAGER_STORE[name]

    _MANAGER_STORE[name] = manager = ComponentManager(name)

    if "." in name:
        root, _ = name.rsplit(".", 1)
        parent = get_manager(root)
        parent.children.add(manager)

    return manager


def check_manager(name: str) -> bool:
    """Check if a manager with the provided name exists.

    .. note::
        Unlike :func:`get_manager`, this function will not create a manager
        if the provided name does not exist.

    Parameters
    ----------
    name:
        The name to check.

    Returns
    -------
    :class:`bool`
        Whether a manager with the provided name exists.

    """
    return name in _MANAGER_STORE



"""Implementation of component base classes.

To easily integrate with disnake-compass, it is recommended to inherit
from any of these base classes. In any case, it is very much recommended to at
least use the `ComponentMeta` metaclass. Without this, a lot of internal
functionality will have to be manually re-implemented.
"""

from __future__ import annotations

import sys
import typing

import attrs
import typing_extensions

from disnake_compass import fields as fields
from disnake_compass.api import component as component_api
from disnake_compass.api import parser as parser_api
from disnake_compass.impl import factory as factory_impl
from disnake_compass.impl import parser as parser_impl

if typing.TYPE_CHECKING:
    import disnake

__all__: typing.Sequence[str] = ("ComponentBase",)


_T = typing.TypeVar("_T")

MaybeCoroutine: typing.TypeAlias = _T | typing.Coroutine[None, None, _T]
_AnyAttr: typing_extensions.TypeAlias = "attrs.Attribute[typing.Any]"


def _is_attrs_pass(namespace: dict[str, typing.Any]) -> bool:
    """Check if attrs has already influenced the class' namespace.

    Note that we check the namespace instead of using `attrs.has`, because
    `attrs.has` would always return `True` for a class inheriting an attrs class,
    and we specifically need to distinguish between the two passes inside
    `ComponentMeta.__new__`.
    """
    return namespace.get("__attrs_attrs__") is not None


def _determine_parser(
    attribute: _AnyAttr,
    overwrite: _AnyAttr | None,
    *,
    required: bool = True,
) -> parser_api.Parser[typing.Any] | None:
    parser = fields.get_parser(attribute)
    if parser:
        return parser

    if overwrite:
        parser = fields.get_parser(overwrite)
        if parser:
            return parser

    if required:
        return parser_impl.get_parser(attribute.type or str)

    return None


def _eval_type(cls: type, annotation: typing.Any) -> typing.Any:  # noqa: ANN401
    # Get the module globals in which the class was defined. This is the most
    # probable candidate in which to find the type annotations' definitions.
    #
    # For the most part, this should be safe. Conflicts where e.g. a component
    # inheriting from RichButton but not defining _AnyEmoji in their own module
    # are safe, because the type has already been passed through this function
    # when the RichButton class was initially created.
    cls_globals = sys.modules[cls.__module__].__dict__

    if isinstance(annotation, str):
        annotation = typing.ForwardRef(annotation, is_argument=False)

    # Evaluate the typehint with the provided globals.
    return typing._eval_type(annotation, cls_globals, None)  # pyright: ignore  # noqa: PGH003, SLF001


def _assert_valid_overwrite(attribute: _AnyAttr, overwrite: _AnyAttr) -> None:
    if fields.FieldMetadata.FIELDTYPE not in overwrite.metadata:
        return

    # The field was defined using fields.field / fields.internal / etc.
    # Ensure the overwrite matches the original field type.

    attribute_type = fields.get_field_type(overwrite)
    overwrite_type = fields.get_field_type(attribute)

    if overwrite_type is not attribute_type:
        new = (attribute_type.name or "<unknown>").lower()
        old = (overwrite_type.name or "<unknown>").lower()
        msg = (
            f"Invalid field override. Field {overwrite.name} is defined as"
            f" a(n) {new} field, but was overwritten as a(n) {old} field."
        )
        raise TypeError(msg)


def _is_custom_id_field(field: _AnyAttr) -> bool:
    return fields.get_field_type(field, fields.FieldType.CUSTOM_ID) is fields.FieldType.CUSTOM_ID


def _field_transformer(
    cls: type,
    attributes: list[_AnyAttr],
) -> list[_AnyAttr]:
    super_attributes: dict[str, _AnyAttr] = (
        {field.name: field for field in fields.get_fields(cls)} if attrs.has(cls) else {}
    )

    finalised_attributes: list[_AnyAttr] = []
    for attribute in attributes:
        super_attribute = super_attributes.get(attribute.name)

        # Ensure all forward-references are evaluated.
        evolved = attribute.evolve(type=_eval_type(cls, attribute.type))

        if super_attribute:
            # This field overwrites a pre-existing field. Merge metadata and
            # ensure the parser isn't overwritten if the new value is None.
            _assert_valid_overwrite(super_attribute, attribute)
            metadata = {**super_attribute.metadata, **evolved.metadata}

        else:
            # Not an overwrite, ensure the fieldtype is set to CUSTOM_ID if not
            # already provided.
            metadata = {
                fields.FieldMetadata.FIELDTYPE: fields.FieldType.CUSTOM_ID,
                **evolved.metadata,
            }

        # TODO: Make copy of parser instead of using the same instance
        metadata[fields.FieldMetadata.PARSER] = _determine_parser(
            evolved,
            super_attribute,
            # Fields only need a parser if
            # - The component is concrete,
            # - The field is a custom-id field.
            # In case of an overwrite, we check the field type of the super-field.
            required=(
                not typing_extensions.is_protocol(cls)
                and _is_custom_id_field(super_attribute or attribute)
            ),
        )

        # Apply finalised metadata.
        finalised_attributes.append(evolved.evolve(metadata=metadata))

    return finalised_attributes


@typing_extensions.dataclass_transform(
    kw_only_default=True,
    field_specifiers=(fields.field, fields.internal),
)
class ComponentMeta(typing._ProtocolMeta):  # pyright: ignore[reportPrivateUsage]  # noqa: SLF001
    """Metaclass for all disnake-compass component types.

    It is **highly** recommended to use this metaclass for any class that
    should interface with the components api exposed by
    `disnake-compass`.

    This metaclass handles :mod:`attrs` class generation, custom id completion,
    interfacing with component managers, parser and factory generation, and
    automatic slotting.
    """

    # HACK: Pyright doesn't like this but it does seem to work with typechecking
    #       down the line. I might change this later (e.g. define it on
    #       BaseComponent instead, but that comes with its own challenges).
    factory: component_api.ComponentFactory[typing_extensions.Self]  # pyright: ignore[reportGeneralTypeIssues, reportUnknownMemberType]

    def __new__(  # noqa: PYI034
        metacls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, typing.Any],
    ) -> ComponentMeta:
        # NOTE: This is run twice for each new class; once for the actual class
        #       definition, and once more by attrs.define(). We ensure we only
        #       run the full class creation logic once.

        # Set slots if attrs hasn't already done so.
        namespace.setdefault("__slots__", ())

        cls = typing.cast(
            "type[ComponentBase]",
            super().__new__(metacls, name, bases, namespace),
        )

        # If this is attrs' pass, return immediately after it has worked its magic.
        if _is_attrs_pass(namespace):
            return cls

        cls = attrs.define(
            cls,
            slots=True,
            kw_only=True,
            field_transformer=_field_transformer,
        )

        if typing_extensions.is_protocol(cls):
            cls.factory = factory_impl.NoopFactory.from_component(cls)
            return cls

        cls.factory = factory_impl.ComponentFactory.from_component(cls)
        return cls


@typing.runtime_checkable
class ComponentBase(
    component_api.RichComponent,
    typing.Protocol,
    metaclass=ComponentMeta,
):
    """Overarching base class for any kind of component."""

    _parent: typing.ClassVar[type[typing.Any] | None] = None
    manager: typing.ClassVar[component_api.ComponentManager | None] = None
    """The manager to which this component is registered.

    Defaults to :obj:`None` if this component is not registered to any manager.
    """

    factory = factory_impl.NoopFactory()
    r"""Factory type that builds instances of this class.

    Since component base classes can be declared as :class:`~typing.Protocol`\s
    and protocols cannot be instantiated, this attribute defaults to a
    :class:`~NoopFactory`. In case a concrete component subclass is created, a matching
    :class:`~ComponentFactory` is automatically generated instead.
    """

    async def as_ui_component(self) -> disnake.ui.WrappedComponent:  # noqa: D102
        # <<Docstring inherited from component_api.RichComponent>>
        ...

    async def make_custom_id(self) -> str:
        """Make a custom id from this component given its current state.

        The generated custom id will contain the full state of the component,
        such that it be used to entirely reconstruct the component later.

        Because parameter to string conversion supports asynchronous callbacks,
        this has to be an async function instead of e.g. a property.

        .. note::
            As the logic for translating a component to and from a custom id
            resides inside the component manager, the component *must* be
            registered to a manager to use this method.

        Returns
        -------
        str:
            The custom id representing the full state of this component.

        """
        if not self.manager:
            message = (
                "A component must be registered to a manager to create a custom"
                "id. Please register this component to a manager before trying"
                "to create a custom id for it."
            )
            raise RuntimeError(message)

        return await self.manager.make_custom_id(self)

    async def callback(  # pyright: ignore[reportIncompatibleMethodOverride]  # noqa: D102
        self,
        inter: disnake.MessageInteraction[disnake.Client],
        /,
    ) -> None:
        # <<docstring inherited from component_api.RichButton>>

        # NOTE: We narrow the interaction type down to a disnake.MessageInteraction
        #       here. This isn't typesafe, but it's just cleaner for the user.
        ...



"""Sentinels for omissible parameters."""

import enum
import typing

import typing_extensions

__all__: typing.Sequence[str] = ("Omissible", "Omitted", "OmittedNoneOr", "is_omitted")


class OmittedType(enum.Enum):
    """Sentinel type for omissible parameters."""

    Omitted = enum.auto()

    def __bool__(self) -> typing.Literal[False]:
        return False


Omitted: typing.Final[typing.Literal[OmittedType.Omitted]] = OmittedType.Omitted
"""Sentinel value for omissible parameters."""

_T = typing.TypeVar("_T")
Omissible: typing.TypeAlias = OmittedType | _T
OmittedNoneOr: typing.TypeAlias = Omissible[_T] | None


def is_omitted(obj: Omissible[_T]) -> typing_extensions.TypeIs[OmittedType]:
    """Check whether a value was omitted."""
    return obj is Omitted



"""Parser implementations for basic disnake snowflake types."""

from __future__ import annotations

import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers

__all__: typing.Sequence[str] = ("SnowflakeParser",)


@parser_base.register_parser_for(disnake.abc.Snowflake, disnake.Object)
@attrs.define(slots=True)
class SnowflakeParser(parser_base.Parser[disnake.abc.Snowflake]):
    r"""Parser implementation for :class:`disnake.abc.Snowflake`\s.

    .. note::
        As snowflakes are abstact, :meth:`loads` returns a
        :class:`disnake.Object` instead.

    Parameters
    ----------
    int_parser:
        The :class:`IntParser` to use internally for this parser.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """

    async def loads(self, argument: str, /) -> disnake.Object:
        """Load a snowflake from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a floating point number.

        """
        return disnake.Object(await self.int_parser.loads(argument))

    async def dumps(self, argument: disnake.abc.Snowflake, /) -> str:
        """Dump a snowflake into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)



"""Protocols for parser types."""

from __future__ import annotations

import typing

import typing_extensions

__all__: typing.Sequence[str] = ("Parser",)


ParserType = typing_extensions.TypeVar(
    "ParserType",
    default=typing.Any,
    infer_variance=True,
)
"""A typevar denoting the type of the parser.

A parser of a given type takes (any subclass of) that type as argument to
:meth:`.Parser.dumps`, and returns (any subclass of) that type from
:meth:`.Parser.loads`.
"""


class Parser(typing.Protocol[ParserType]):
    """The baseline protocol for any kind of parser.

    Any and all parser types must implement this protocol in order to be
    properly handled by disnake-compass.
    """

    __slots__: typing.Sequence[str] = ()

    @classmethod
    def default(cls, target_type: type[ParserType], /) -> typing_extensions.Self:
        """Return the default implementation of this parser type.

        By default, this will just create the parser class with no arguments,
        but this can be overwritten on child classes for customised behaviour.

        Parameters
        ----------
        target_type:
            The exact type that this parser should be created for

        Returns
        -------
        Parser:
            The default parser instance for this parser type.

        """
        ...

    async def loads(self, argument: str, /) -> ParserType:
        r"""Load a value from a string and apply the necessary conversion logic.

        Any errors raised inside this method remain unmodified, and should be
        handled externally.

        Parameters
        ----------
        argument:
            The argument to parse into the desired type.

        Returns
        -------
        :data:`.ParserType`:
            In case the parser method was sync, the parsed result is returned
            as-is.
        :class:`~typing.Coroutine`\[:data:`.ParserType`]:
            In case the parser method was async, the parser naturally returns a
            coroutine. Awaiting this coroutine returns the parser result.

        """
        ...

    async def dumps(self, argument: ParserType, /) -> str:
        r"""Dump a value from a given type and convert it to a string.

        In most cases it is imperative to ensure that this is done in a
        reversible way, such that calling :meth:`loads` on the result of this
        function returns the original input. For example:

        .. code-block:: python3

            >>> parser = IntParser()
            >>> input_str = "1"
            >>> parsed_int = parser.loads(input_str)
            >>> dumped_int = parser.dumps(parsed_int)
            >>> input_str == dumped_int
            True

        Any errors raised inside this method remain unmodified, and should be
        handled externally.

        Parameters
        ----------
        argument:
            The argument to parse into the desired type.

        Returns
        -------
        :class:`str`:
            In case the parser method was sync, the resulting dumped argument.
        :class:`~typing.Coroutine`\[:class:`str`]:
            In case the parser method was async, the parser naturally returns a
            coroutine. Awaiting this coroutine returns the dumped argument.

        """
        ...



"""Parser implementations for disnake channel types."""

from __future__ import annotations

import contextlib
import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers
from disnake_compass.internal import di

__all__: typing.Sequence[str] = (
    "CategoryParser",
    "DMChannelParser",
    "ForumChannelParser",
    "GroupChannelParser",
    "GuildChannelParser",
    "NewsChannelParser",
    "PartialMessageableParser",
    "PrivateChannelParser",
    "StageChannelParser",
    "TextChannelParser",
    "ThreadParser",
    "VoiceChannelParser",
)


_AnyChannel: typing.TypeAlias = (
    disnake.abc.GuildChannel | disnake.abc.PrivateChannel | disnake.Thread
)
_ChannelT = typing.TypeVar("_ChannelT", bound=_AnyChannel)


# NOTE: Making these protocols messes with documentation of all subclasses'
#       __init__ methods.
@attrs.define(slots=True, init=False)
class ChannelParserBase(parser_base.Parser[_ChannelT]):
    r"""Base class for synchronous parser types with support for channels.

    .. note::
        This class cannot be instantiated.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    parser_type: type[_ChannelT]  # NOTE: Intentionally undocumented.
    int_parser: builtins_parsers.IntParser
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default channel parser will also return compressed results.
    """
    allow_api_requests: bool
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    def __init__(
        self,
        int_parser: builtins_parsers.IntParser | None = None,
        *,
        allow_api_requests: bool = True,
    ) -> None:
        if type(self) is ChannelParserBase:
            msg = "'GetChannelParserBase' is a base class and should not be instantiated directly."
            raise TypeError(msg)

        self.int_parser = int_parser or builtins_parsers.IntParser.default(int)
        self.allow_api_requests = allow_api_requests

    async def loads(self, argument: str, /) -> _ChannelT:
        """Load a channel from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a channel.

            This always matches the channel type of the parser.

        Raises
        ------
        :class:`LookupError`:
            A channel with the id stored in the ``argument`` could not be found.
        :class:`TypeError`:
            A channel with the id stored in the ``argument`` was found, but it
            was of an incorrect channel type.

        """
        channel_id = await self.int_parser.loads(argument)

        maybe_channel = (
            di.resolve_dependency(disnake.abc.GuildChannel, None)
            or di.resolve_dependency(disnake.Thread, None)
            or di.resolve_dependency(disnake.abc.PrivateChannel, None)
        )
        if maybe_channel and isinstance(maybe_channel, self.parser_type):
            return maybe_channel

        maybe_guild = di.resolve_dependency(disnake.Guild, None)
        if maybe_guild:
            maybe_channel = maybe_guild.get_channel(channel_id)
            if maybe_channel and isinstance(maybe_channel, self.parser_type):
                return maybe_channel

        maybe_client = di.resolve_dependency(disnake.Client, None)
        if maybe_client:
            maybe_channel = maybe_client.get_channel(channel_id)
            if maybe_channel and isinstance(maybe_channel, self.parser_type):
                return maybe_channel

            if self.allow_api_requests:
                with contextlib.suppress(disnake.HTTPException):
                    maybe_channel = await maybe_client.fetch_channel(channel_id)

                if isinstance(maybe_channel, self.parser_type):
                    return maybe_channel

        if maybe_channel is None:
            msg = f"Could not find a channel with id {argument!r}."
            raise LookupError(msg)

        msg = (
            f"Found a channel of type {type(maybe_channel).__name__!r} for id"
            f" {argument!r}, expected type {self.parser_type.__name__!r}."
        )
        raise TypeError(msg)

    async def dumps(self, argument: _ChannelT, /) -> str:
        """Dump a channel into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)


# ABSTRACT


@parser_base.register_parser_for(disnake.abc.GuildChannel)
@attrs.define(slots=True)
class GuildChannelParser(ChannelParserBase[disnake.abc.GuildChannel]):
    r"""Parser type with support for guild channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.abc.GuildChannel


@parser_base.register_parser_for(disnake.abc.PrivateChannel)
@attrs.define(slots=True)
class PrivateChannelParser(ChannelParserBase[disnake.abc.PrivateChannel]):
    r"""Parser type with support for private channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.abc.PrivateChannel  # pyright: ignore[reportAssignmentType]


# ASYNC PRIVATE


@parser_base.register_parser_for(disnake.DMChannel)
@attrs.define(slots=True)
class DMChannelParser(ChannelParserBase[disnake.DMChannel]):
    r"""Parser type with support for DM channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.DMChannel


@parser_base.register_parser_for(disnake.GroupChannel)
@attrs.define(slots=True)
class GroupChannelParser(ChannelParserBase[disnake.GroupChannel]):
    r"""Parser type with support for group channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.GroupChannel


# ASYNC GUILD


@parser_base.register_parser_for(disnake.ForumChannel)
@attrs.define(slots=True)
class ForumChannelParser(ChannelParserBase[disnake.ForumChannel]):
    r"""Parser type with support for forum channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.ForumChannel


@parser_base.register_parser_for(disnake.NewsChannel)
@attrs.define(slots=True)
class NewsChannelParser(ChannelParserBase[disnake.NewsChannel]):
    r"""Parser type with support for news channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.NewsChannel


@parser_base.register_parser_for(disnake.VoiceChannel)
@attrs.define(slots=True)
class VoiceChannelParser(ChannelParserBase[disnake.VoiceChannel]):
    r"""Parser type with support for voice channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.VoiceChannel


@parser_base.register_parser_for(disnake.StageChannel)
@attrs.define(slots=True)
class StageChannelParser(ChannelParserBase[disnake.StageChannel]):
    r"""Parser type with support for stage channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.StageChannel


@parser_base.register_parser_for(disnake.TextChannel)
@attrs.define(slots=True)
class TextChannelParser(ChannelParserBase[disnake.TextChannel]):
    r"""Parser type with support for text channels.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.TextChannel


@parser_base.register_parser_for(disnake.Thread)
@attrs.define(slots=True)
class ThreadParser(ChannelParserBase[disnake.Thread]):
    r"""Parser type with support for threads.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.Thread


@parser_base.register_parser_for(disnake.CategoryChannel)
@attrs.define(slots=True)
class CategoryParser(ChannelParserBase[disnake.CategoryChannel]):
    r"""Parser type with support for categories.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    parser_type = disnake.CategoryChannel


@parser_base.register_parser_for(disnake.PartialMessageable)
@attrs.define(slots=True)
class PartialMessageableParser(parser_base.Parser[disnake.PartialMessageable]):
    r"""Parser type with support for partial messageables.

    Parameters
    ----------
    channel_type:
        The channel type to use for :class:`disnake.PartialMessageable`\s
        created by this class.

        Defaults to ``None``.
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.

    """

    channel_type: disnake.ChannelType | None
    r"""The channel type to use for :class:`disnake.PartialMessageable`\s
    created by this class.

    This determines which operations are valid on the partial messageables.
    """
    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default channel parser will also return compressed results.
    """

    async def loads(self, argument: str, /) -> disnake.PartialMessageable:
        """Load a partial messageable from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a channel.

            This always matches the channel type of the parser.

        """
        client = di.resolve_dependency(disnake.Client)
        return client.get_partial_messageable(int(argument), type=self.channel_type)

    async def dumps(self, argument: disnake.PartialMessageable, /) -> str:
        """Dump a partial messageable into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)



"""Standard implementation of the overarching component factory type."""

from __future__ import annotations

import types
import typing

import attrs

from disnake_compass import fields
from disnake_compass.api import component as component_api
from disnake_compass.api import parser as parser_api
from disnake_compass.impl.parser import base as parser_base

if typing.TYPE_CHECKING:
    import typing_extensions

__all__: typing.Sequence[str] = ("ComponentFactory",)


ParserMapping = typing.Mapping[str, parser_api.Parser[typing.Any]]


@attrs.define(slots=True)
class ComponentFactory(
    component_api.ComponentFactory[component_api.ComponentT],
    typing.Generic[component_api.ComponentT],
):
    """Implementation of the overarching component factory type.

    A component factory holds information about all the custom id fields of a
    component, and contains that component's parsers. In most situations, a
    component factory can simply be created using :meth:`from_component`.
    """

    parsers: ParserMapping = attrs.field(converter=types.MappingProxyType)  # pyright: ignore[reportGeneralTypeIssues]
    """A mapping of custom id field name to that field's parser."""
    component: type[component_api.ComponentT]
    """The component type that this factory builds."""

    @classmethod
    def from_component(  # noqa: D102
        cls,
        component: type[component_api.RichComponent],
    ) -> typing_extensions.Self:
        # <<docstring inherited from api.components.ComponentFactory>>
        parser: parser_api.Parser[typing.Any] | None

        parsers: dict[str, parser_api.Parser[typing.Any]] = {}
        for field in fields.get_fields(component, kind=fields.FieldType.CUSTOM_ID):
            parser = fields.get_parser(field)

            if not parser:
                parser_type = field.type or str
                parser = parser_base.get_parser(parser_type).default(parser_type)

            parsers[field.name] = parser

        return cls(
            parsers,
            typing.cast(type[component_api.ComponentT], component),
        )

    async def load_params(  # noqa: D102
        self,
        params: typing.Sequence[str],
    ) -> typing.Mapping[str, object]:
        # <<docstring inherited from api.components.ComponentFactory>>
        return {
            param: await self.parsers[param].loads(value)
            for param, value in zip(self.parsers, params, strict=True)
            if value  # TODO: Check this, I think this is wrong.
        }

    async def dump_params(  # noqa: D102
        self,
        component: component_api.ComponentT,
    ) -> typing.Mapping[str, str]:
        # <<docstring inherited from api.components.ComponentFactory>>

        return {
            field: await self.parsers[field].dumps(getattr(component, field))
            for field in self.parsers
        }

    async def build_component(  # noqa: D102
        self,
        params: typing.Sequence[str],
        component_params: typing.Mapping[str, object] | None = None,
    ) -> component_api.ComponentT:
        # <<docstring inherited from api.components.ComponentFactory>>

        parsed = await self.load_params(params)
        return self.component(**parsed, **(component_params or {}))


class NoopFactory(component_api.ComponentFactory[typing.Any]):
    """Factory class to make component protocols typesafe.

    Since component protocols cannot be instantiated, building a factory with
    parsers for them does not make sense. Instead, they will receive one of
    these to remain typesafe. Any operation on a NoopFactory will raise
    :class:`NotImplementedError`.
    """

    __slots__: typing.Sequence[str] = ()
    __instance: typing.ClassVar[typing_extensions.Self | None] = None

    def __new__(cls) -> typing_extensions.Self:
        if cls.__instance is not None:
            return cls.__instance  # pyright: ignore[reportReturnType]

        cls.__instance = self = super().__new__(cls)
        return self

    @classmethod
    def from_component(
        cls,
        _: type[component_api.RichComponent],
    ) -> typing_extensions.Self:
        # <<docstring inherited from api.components.ComponentFactory>>

        return cls()

    async def load_params(self, *_: object) -> typing.NoReturn:
        # <<docstring inherited from api.components.ComponentFactory>>

        raise NotImplementedError

    async def dump_params(self, *_: object) -> typing.NoReturn:
        # <<docstring inherited from api.components.ComponentFactory>>

        raise NotImplementedError

    async def build_component(
        self,
        params: typing.Sequence[str],
        component_params: typing.Mapping[str, object] | None = None,
    ) -> typing.NoReturn:
        # <<docstring inherited from api.components.ComponentFactory>>

        raise NotImplementedError

    def __repr__(self) -> str:
        return "<NoopFactory>"



"""Parser implementations for disnake message types."""

from __future__ import annotations

import contextlib
import typing

import attrs
import disnake

from disnake_compass.impl.parser import base as parser_base
from disnake_compass.impl.parser import builtins as builtins_parsers
from disnake_compass.internal import di

__all__: typing.Sequence[str] = ("MessageParser", "PartialMessageParser")

AnyChannel: typing.TypeAlias = (
    disnake.TextChannel
    | disnake.Thread
    | disnake.VoiceChannel
    | disnake.DMChannel
    | disnake.PartialMessageable
)


@typing.runtime_checkable
class SupportsGetPartialMessage(typing.Protocol):
    def get_partial_message(self, message_id: int, /) -> disnake.PartialMessage: ...


@parser_base.register_parser_for(disnake.PartialMessage)
@attrs.define(slots=True)
class PartialMessageParser(parser_base.Parser[disnake.PartialMessage]):
    r"""Parser type with support for partial messages.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    channel:
        The channel in which to make the partial message.

        Defaults to ``None``. If left to be ``None``, :meth:`loads` will
        attempt to get a channel from the ``source``.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    channel: SupportsGetPartialMessage | None = attrs.field(default=None, kw_only=True)
    """The channel in which to make the partial message."""

    async def loads(self, argument: str, /) -> disnake.PartialMessage:
        """Load a partial message from a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a partial message.

        Raises
        ------
        :class:`RuntimeError`:
            :attr:`channel` was not set, and no channel could be obtained from
            the ``source``.

        """
        channel = self.channel or di.resolve_dependency(SupportsGetPartialMessage, None)
        if not channel:
            msg = (
                "A channel must be provided either through self.channel or"
                " dependency injection, got neither."
            )
            raise RuntimeError(msg)

        return channel.get_partial_message(await self.int_parser.loads(argument))

    async def dumps(self, argument: disnake.PartialMessage, /) -> str:
        """Dump a partial message into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)


@parser_base.register_parser_for(disnake.Message)
@attrs.define(slots=True)
class MessageParser(parser_base.Parser[disnake.Message]):
    r"""Asynchronous parser type with support for messages.

    .. warning::
        This parser can make API requests.

    Parameters
    ----------
    int_parser:
        The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
        internally for this parser.
    allow_api_requests:
        Whether or not to allow this parser to make API requests.

    """

    int_parser: builtins_parsers.IntParser = attrs.field(
        factory=lambda: builtins_parsers.IntParser.default(int),
    )
    """The :class:`~disnake_compass.impl.parser.builtins.IntParser` to use
    internally for this parser.

    Since the default integer parser uses base-36 to "compress" numbers, the
    default guild parser will also return compressed results.
    """
    allow_api_requests: bool = attrs.field(default=True, kw_only=True)
    """Whether or not to allow this parser to make API requests.

    Parsers will always try getting a result from cache first.
    """

    async def loads(self, argument: str, /) -> disnake.Message:
        """Load a message from a string.

        This uses the underlying :attr:`int_parser`.

        This method first tries to get the message from cache. If this fails,
        it will try to fetch the message instead.

        .. warning::
            This method can make API requests.

        Parameters
        ----------
        argument:
            The value that is to be loaded into a message.

        Raises
        ------
        :class:`LookupError`:
            A message with the id stored in the ``argument`` could not be found.

        """
        message_id = await self.int_parser.loads(argument)

        maybe_message = di.resolve_dependency(disnake.Message, None)
        if maybe_message and maybe_message.id == message_id:
            return maybe_message

        maybe_client = di.resolve_dependency(disnake.Client, None)
        if maybe_client:
            message = maybe_client.get_message(message_id)
            if message:
                return message

        if self.allow_api_requests:
            maybe_messageable = di.resolve_dependency(disnake.abc.Messageable, None)
            if maybe_messageable:
                with contextlib.suppress(disnake.HTTPException):
                    return await maybe_messageable.fetch_message(message_id)

        msg = f"Could not find a message with id {argument!r}."
        raise LookupError(msg)

    async def dumps(self, argument: disnake.Message, /) -> str:
        """Dump a message into a string.

        This uses the underlying :attr:`int_parser`.

        Parameters
        ----------
        argument:
            The value that is to be dumped.

        """
        return await self.int_parser.dumps(argument.id)
